from typing import Any, Dict


def merge_dicts(dict1: Dict, dict2: Dict) -> Dict:
    if dict1 and not dict2:
        return dict(dict1)
    elif dict2 and not dict1:
        return dict(dict2)
    elif not dict1 and not dict2:
        return {}

    context = dict(dict1)
    for k, v2 in dict2.items():
        v1 = context.get(k)
        if not context.get(k):
            context[k] = v2
        elif isinstance(v1, list) and isinstance(v2, list):
            context[k] = v1 + v2
        elif isinstance(v1, dict) and isinstance(v2, dict):
            context[k] = merge_dicts(v1, v2)
        else:
            context[k] = v2

    return context


def get_item_by_path(dict: Dict, path: str, default: Any = None) -> Any:
    if "." not in path:
        raise KeyError("Key path must contain '.' ")
    doted_paths = path.split(".")
    item = dict.get(doted_paths[0], {})
    for i in range(1, len(doted_paths)):
        if i == len(doted_paths) - 1:
            default_value = default
        else:
            default_value = {}
        if isinstance(item, Dict):
            item = item.get(doted_paths[i], default_value)
        else:
            raise ValueError("Item at key path {} is not a Dict".format(".".join(doted_paths[:i])))
    return item



import asyncio
import functools
import inspect
import ipaddress
import logging
import os
import pathlib
import platform
import re
import time
import uuid
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, SupportsInt, Tuple, Union, cast

import yarl
from aiohttp import WSMsgType
from aiohttp import __version__ as aiohttp_version
from aiohttp import hdrs, web, web_protocol, web_server, web_urldispatcher
from aiohttp.helpers import BasicAuth
from aiohttp.http import HttpVersion
from aiohttp.streams import EofStream
from aiohttp.web_fileresponse import FileResponse
from multidict import CIMultiDict, CIMultiDictProxy

from tomodachi.helpers.dict import merge_dicts
from tomodachi.helpers.execution_context import (
    decrease_execution_context_value,
    increase_execution_context_value,
    set_execution_context,
)
from tomodachi.helpers.middleware import execute_middlewares
from tomodachi.invoker import Invoker

http_logger = logging.getLogger("transport.http")


# Should be implemented as lazy load instead
class ColoramaCache:
    _is_colorama_installed: Optional[bool] = None
    _colorama: Any = None


class HttpException(Exception):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._log_level = kwargs.get("log_level") if kwargs and kwargs.get("log_level") else "INFO"


class RequestHandler(web_protocol.RequestHandler):
    __slots__ = (
        *web_protocol.RequestHandler.__slots__,
        "_server_header",
        "_access_log",
        "_connection_start_time",
        "_keepalive",
    )

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._server_header = kwargs.pop("server_header", None) if kwargs else None
        self._access_log = kwargs.pop("access_log", None) if kwargs else None

        self._connection_start_time = time.time()

        super().__init__(*args, access_log=None, **kwargs)  # type: ignore

    @staticmethod
    def get_request_ip(request: Any, context: Optional[Dict] = None) -> Optional[str]:
        if request._cache.get("request_ip"):
            return str(request._cache.get("request_ip", ""))

        if request.transport:
            if not context:
                context = {}
            real_ip_header = context.get("options", {}).get("http", {}).get("real_ip_header", "X-Forwarded-For")
            real_ip_from = context.get("options", {}).get("http", {}).get("real_ip_from", [])
            if isinstance(real_ip_from, str):
                real_ip_from = [real_ip_from]

            peername = request.transport.get_extra_info("peername")
            request_ip = None
            if peername:
                request_ip, _ = peername
            if (
                real_ip_header
                and real_ip_from
                and request.headers.get(real_ip_header)
                and request_ip
                and len(real_ip_from)
            ):
                if any([ipaddress.ip_address(request_ip) in ipaddress.ip_network(cidr) for cidr in real_ip_from]):
                    request_ip = request.headers.get(real_ip_header).split(",")[0].strip().split(" ")[0].strip()

            request._cache["request_ip"] = request_ip
            return request_ip

        return None

    @staticmethod
    def colorize_status(text: Optional[Union[str, int]], status: Optional[Union[str, int, bool]] = False) -> str:
        if ColoramaCache._is_colorama_installed is None:
            try:
                import colorama  # noqa  # isort:skip

                ColoramaCache._is_colorama_installed = True
                ColoramaCache._colorama = colorama
            except Exception:
                ColoramaCache._is_colorama_installed = False

        if ColoramaCache._is_colorama_installed is False:
            return str(text) if text else ""

        if status is False:
            status = text
        status_code = str(status) if status else None
        if status_code and not http_logger.handlers:
            output_text = str(text) if text else ""
            color = None

            if status_code == "101":
                color = ColoramaCache._colorama.Fore.CYAN
            elif status_code[0] == "2":
                color = ColoramaCache._colorama.Fore.GREEN
            elif status_code[0] == "3" or status_code == "499":
                color = ColoramaCache._colorama.Fore.YELLOW
            elif status_code[0] == "4":
                color = ColoramaCache._colorama.Fore.RED
            elif status_code[0] == "5":
                color = ColoramaCache._colorama.Fore.WHITE + ColoramaCache._colorama.Back.RED

            if color:
                return "{}{}{}".format(color, output_text, ColoramaCache._colorama.Style.RESET_ALL)
            return output_text

        return str(text) if text else ""

    def handle_error(
        self, request: Any, status: int = 500, exc: Any = None, message: Optional[str] = None
    ) -> web.Response:
        """Handle errors.

        Returns HTTP response with specific status code. Logs additional
        information. It always closes current connection."""
        if self.transport is None:
            # client has been disconnected during writing.
            if self._access_log:
                request_ip = RequestHandler.get_request_ip(request, None)
                version_string = None
                if isinstance(request.version, HttpVersion):
                    version_string = "HTTP/{}.{}".format(request.version.major, request.version.minor)
                http_logger.info(
                    '[{}] [{}] {} {} "{} {}{}{}" - {} "{}" -'.format(
                        RequestHandler.colorize_status("http", 499),
                        RequestHandler.colorize_status(499),
                        request_ip or "",
                        '"{}"'.format(request._cache["auth"].login.replace('"', ""))
                        if request._cache.get("auth") and getattr(request._cache.get("auth"), "login", None)
                        else "-",
                        request.method,
                        request.path,
                        "?{}".format(request.query_string) if request.query_string else "",
                        " {}".format(version_string) if version_string else "",
                        request.content_length if request.content_length is not None else "-",
                        request.headers.get("User-Agent", "").replace('"', ""),
                    )
                )

        headers: CIMultiDict = CIMultiDict({})
        headers[hdrs.CONTENT_TYPE] = "text/plain; charset=utf-8"

        msg = "" if status == 500 or not message else message

        headers[hdrs.CONTENT_LENGTH] = str(len(msg))
        headers[hdrs.SERVER] = self._server_header or ""

        if isinstance(request.version, HttpVersion) and (request.version.major, request.version.minor) in (
            (1, 0),
            (1, 1),
        ):
            headers[hdrs.CONNECTION] = "close"

        resp: web.Response = web.Response(status=status, text=msg, headers=headers)
        resp.force_close()

        # some data already got sent, connection is broken
        if request.writer.output_size > 0 or self.transport is None:
            self.force_close()
        elif self.transport is not None:
            request_ip = RequestHandler.get_request_ip(request, None)
            if not request_ip:
                peername = request.transport.get_extra_info("peername")
                if peername:
                    request_ip, _ = peername
            if self._access_log:
                http_logger.info(
                    '[{}] [{}] {} {} "INVALID" {} - "" -'.format(
                        RequestHandler.colorize_status("http", status),
                        RequestHandler.colorize_status(status),
                        request_ip or "",
                        '"{}"'.format(request._cache["auth"].login.replace('"', ""))
                        if request._cache.get("auth") and getattr(request._cache.get("auth"), "login", None)
                        else "-",
                        len(msg),
                    )
                )

        return resp


class Server(web_server.Server):
    __slots__ = (
        "_loop",
        "_connections",
        "_kwargs",
        "requests_count",
        "request_handler",
        "request_factory",
        "_server_header",
        "_access_log",
    )

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._server_header = kwargs.pop("server_header", None) if kwargs else None
        self._access_log = kwargs.pop("access_log", None) if kwargs else None

        super().__init__(*args, **kwargs)

    def __call__(self) -> RequestHandler:
        return RequestHandler(
            self, loop=self._loop, server_header=self._server_header, access_log=self._access_log, **self._kwargs
        )


class DynamicResource(web_urldispatcher.DynamicResource):
    def __init__(self, pattern: Any, *, name: Optional[str] = None) -> None:
        self._routes: List = []
        self._name = name
        self._pattern = pattern
        self._formatter = ""


class Response(object):
    __slots__ = ("_body", "_status", "_reason", "_headers", "content_type", "charset", "missing_content_type")

    def __init__(
        self,
        *,
        body: Optional[Union[bytes, str]] = None,
        status: int = 200,
        reason: Optional[str] = None,
        headers: Optional[Union[Dict, CIMultiDict, CIMultiDictProxy]] = None,
        content_type: Optional[str] = None,
        charset: Optional[str] = None,
    ) -> None:
        if headers is None:
            headers = CIMultiDict()
        elif not isinstance(headers, (CIMultiDict, CIMultiDictProxy)):
            headers = CIMultiDict(headers)

        self._body = body
        self._status = status
        self._reason = reason
        self._headers = headers
        self.content_type = content_type if hdrs.CONTENT_TYPE not in headers else None
        self.charset = charset if hdrs.CONTENT_TYPE not in headers else None

        self.missing_content_type = hdrs.CONTENT_TYPE not in headers and not content_type and not charset

    def get_aiohttp_response(
        self, context: Dict, default_charset: Optional[str] = None, default_content_type: Optional[str] = None
    ) -> web.Response:
        if self.missing_content_type:
            self.charset = default_charset
            self.content_type = default_content_type

        charset = self.charset
        if hdrs.CONTENT_TYPE in self._headers and ";" in self._headers[hdrs.CONTENT_TYPE]:
            try:
                charset = (
                    str([v for v in self._headers[hdrs.CONTENT_TYPE].split(";") if "charset=" in v][0])
                    .replace("charset=", "")
                    .strip()
                )
            except IndexError:
                pass
        elif hdrs.CONTENT_TYPE in self._headers and ";" not in self._headers[hdrs.CONTENT_TYPE]:
            charset = None

        if self._body and not isinstance(self._body, bytes) and charset:
            body = self._body
            try:
                body_value = body.encode(charset.lower())
            except (ValueError, LookupError, UnicodeEncodeError) as e:
                logging.getLogger("exception").exception("Uncaught exception: {}".format(str(e)))
                raise web.HTTPInternalServerError() from e
        elif self._body:
            body_value = self._body.encode() if not isinstance(self._body, bytes) else self._body
        else:
            body_value = b""

        response: web.Response = web.Response(
            body=body_value,
            status=self._status,
            reason=self._reason,
            headers=self._headers,
            content_type=self.content_type,
            charset=self.charset,
        )
        return response


class HttpTransport(Invoker):
    server_port_mapping: Dict[Any, str] = {}

    @classmethod
    async def request_handler(
        cls,
        obj: Any,
        context: Dict,
        func: Any,
        method: Union[str, List[str], Tuple[str, ...]],
        url: str,
        *,
        ignore_logging: Union[bool, List[int], Tuple[int, ...]] = False,
        pre_handler_func: Optional[Callable] = None,
    ) -> Any:
        pattern = r"^{}$".format(re.sub(r"\$$", "", re.sub(r"^\^?(.*)$", r"\1", url)))
        compiled_pattern = re.compile(pattern)

        default_content_type = context.get("options", {}).get("http", {}).get("content_type", "text/plain")
        default_charset = context.get("options", {}).get("http", {}).get("charset", "utf-8")

        if default_content_type is not None and ";" in default_content_type:
            # for backwards compability
            try:
                default_charset = (
                    str([v for v in default_content_type.split(";") if "charset=" in v][0])
                    .replace("charset=", "")
                    .strip()
                )
                default_content_type = str([v for v in default_content_type.split(";")][0]).strip()
            except IndexError:
                pass

        values = inspect.getfullargspec(func)
        original_kwargs = (
            {k: values.defaults[i] for i, k in enumerate(values.args[len(values.args) - len(values.defaults) :])}
            if values.defaults
            else {}
        )

        middlewares = context.get("http_middleware", [])

        async def handler(request: web.Request) -> Union[web.Response, web.FileResponse]:
            kwargs = dict(original_kwargs)
            if "(" in pattern:
                result = compiled_pattern.match(request.path)
                if result:
                    for k, v in result.groupdict().items():
                        kwargs[k] = v

            @functools.wraps(func)
            async def routine_func(
                *a: Any, **kw: Any
            ) -> Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response]:
                routine = func(*(obj, request, *a), **merge_dicts(kwargs, kw))
                return_value: Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response] = (
                    (await routine) if inspect.isawaitable(routine) else routine
                )
                return return_value

            if not context.get("_http_accept_new_requests"):
                raise web.HTTPServiceUnavailable()

            if pre_handler_func:
                await pre_handler_func(obj, request)

            return_value: Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response]
            if middlewares:
                return_value = await execute_middlewares(func, routine_func, middlewares, *(obj, request))
            else:
                routine = func(obj, request, **kwargs)
                return_value = (await routine) if inspect.isawaitable(routine) else routine

            response = resolve_response_sync(
                return_value,
                request=request,
                context=context,
                default_content_type=default_content_type,
                default_charset=default_charset,
            )
            return response

        context["_http_routes"] = context.get("_http_routes", [])
        route_context = {"ignore_logging": ignore_logging}
        if isinstance(method, list) or isinstance(method, tuple):
            for m in method:
                context["_http_routes"].append((m.upper(), pattern, handler, route_context))
        elif isinstance(method, str):
            context["_http_routes"].append((method.upper(), pattern, handler, route_context))
        else:
            raise Exception("Invalid method '{}' for route".format(str(method)))

        start_func = cls.start_server(obj, context)
        return (await start_func) if start_func else None

    @classmethod
    async def static_request_handler(
        cls,
        obj: Any,
        context: Dict,
        func: Any,
        path: str,
        base_url: str,
        *,
        ignore_logging: Union[bool, List[int], Tuple[int, ...]] = False,
    ) -> Any:
        if "?P<filename>" not in base_url:
            pattern = r"^{}(?P<filename>.+?)$".format(re.sub(r"\$$", "", re.sub(r"^\^?(.*)$", r"\1", base_url)))
        else:
            pattern = r"^{}$".format(re.sub(r"\$$", "", re.sub(r"^\^?(.*)$", r"\1", base_url)))
        compiled_pattern = re.compile(pattern)

        if path in ("", "/"):
            # Hopefully noone wants to do this intentionally, and if anyone accidentally does we'll catch it here
            raise Exception("Invalid path '{}' for static route".format(path))

        if not path.startswith("/"):
            path = "{}/{}".format(os.path.dirname(context.get("context", {}).get("_service_file_path", "")), path)

        if not path.endswith("/"):
            path = "{}/".format(path)

        if os.path.realpath(path) == "/":
            raise Exception("Invalid path '{}' for static route resolves to '/'".format(path))

        async def handler(request: web.Request) -> Union[web.Response, web.FileResponse]:
            normalized_request_path = yarl.URL._normalize_path(request.path)
            if not normalized_request_path.startswith("/"):
                raise web.HTTPNotFound()

            result = compiled_pattern.match(normalized_request_path)
            filename = result.groupdict()["filename"] if result else ""

            basepath = os.path.realpath(path)
            realpath = os.path.realpath("{}/{}".format(basepath, filename))

            try:
                if any(
                    [
                        not realpath,
                        not basepath,
                        realpath == basepath,
                        os.path.commonprefix((realpath, basepath)) != basepath,
                        not os.path.exists(realpath),
                        not os.path.isdir(basepath),
                        basepath == "/",
                        os.path.isdir(realpath),
                    ]
                ):
                    raise web.HTTPNotFound()

                # deepcode ignore PT: Input data to pathlib.Path is sanitized
                pathlib.Path(realpath).open("r")

                response: Union[web.Response, web.FileResponse] = FileResponse(path=realpath, chunk_size=256 * 1024)
                return response
            except PermissionError:
                raise web.HTTPForbidden()

        route_context = {"ignore_logging": ignore_logging}
        context["_http_routes"] = context.get("_http_routes", [])
        context["_http_routes"].append(("GET", pattern, handler, route_context))

        start_func = cls.start_server(obj, context)
        return (await start_func) if start_func else None

    @classmethod
    async def error_handler(cls, obj: Any, context: Dict, func: Any, status_code: int) -> Any:
        default_content_type = context.get("options", {}).get("http", {}).get("content_type", "text/plain")
        default_charset = context.get("options", {}).get("http", {}).get("charset", "utf-8")

        if default_content_type is not None and ";" in default_content_type:
            # for backwards compability
            try:
                default_charset = (
                    str([v for v in default_content_type.split(";") if "charset=" in v][0])
                    .replace("charset=", "")
                    .strip()
                )
                default_content_type = str([v for v in default_content_type.split(";")][0]).strip()
            except IndexError:
                pass

        values = inspect.getfullargspec(func)
        kwargs = (
            {k: values.defaults[i] for i, k in enumerate(values.args[len(values.args) - len(values.defaults) :])}
            if values.defaults
            else {}
        )

        middlewares = context.get("http_middleware", [])

        async def handler(request: web.Request) -> Union[web.Response, web.FileResponse]:
            request._cache["error_status_code"] = status_code

            @functools.wraps(func)
            async def routine_func(
                *a: Any, **kw: Any
            ) -> Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response]:
                routine = func(*(obj, request, *a), **merge_dicts(kwargs, kw))
                return_value: Union[str, bytes, Dict, List, Tuple, web.Response, Response] = (
                    (await routine) if inspect.isawaitable(routine) else routine
                )
                return return_value

            return_value: Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response]
            if middlewares:
                return_value = await execute_middlewares(func, routine_func, middlewares, *(obj, request))
            else:
                routine = func(obj, request, **kwargs)
                return_value = (await routine) if inspect.isawaitable(routine) else routine

            response = resolve_response_sync(
                return_value,
                request=request,
                context=context,
                status_code=status_code,
                default_content_type=default_content_type,
                default_charset=default_charset,
            )
            return response

        context["_http_error_handler"] = context.get("_http_error_handler", {})
        context["_http_error_handler"][int(status_code)] = handler

        start_func = cls.start_server(obj, context)
        return (await start_func) if start_func else None

    @classmethod
    async def websocket_handler(cls, obj: Any, context: Dict, func: Any, url: str) -> Any:
        pattern = r"^{}$".format(re.sub(r"\$$", "", re.sub(r"^\^?(.*)$", r"\1", url)))
        compiled_pattern = re.compile(pattern)

        access_log = context.get("options", {}).get("http", {}).get("access_log", True)

        async def _pre_handler_func(_: Any, request: web.Request) -> None:
            request._cache["is_websocket"] = True
            request._cache["websocket_uuid"] = str(uuid.uuid4())

        values = inspect.getfullargspec(func)
        original_kwargs = (
            {k: values.defaults[i] for i, k in enumerate(values.args[len(values.args) - len(values.defaults) :])}
            if values.defaults
            else {}
        )

        @functools.wraps(func)
        async def _func(obj: Any, request: web.Request, *a: Any, **kw: Any) -> None:
            websocket = web.WebSocketResponse()

            request_ip = RequestHandler.get_request_ip(request, context)
            try:
                await websocket.prepare(request)
            except Exception:
                try:
                    await websocket.close()
                except Exception:
                    pass

                if access_log:
                    http_logger.info(
                        '[{}] {} {} "CANCELLED {}{}" {} "{}" {}'.format(
                            RequestHandler.colorize_status("websocket", 101),
                            request_ip,
                            '"{}"'.format(request._cache["auth"].login.replace('"', ""))
                            if request._cache.get("auth") and getattr(request._cache.get("auth"), "login", None)
                            else "-",
                            request.path,
                            "?{}".format(request.query_string) if request.query_string else "",
                            request._cache.get("websocket_uuid", ""),
                            request.headers.get("User-Agent", "").replace('"', ""),
                            "-",
                        )
                    )

                return

            context["_http_open_websockets"] = context.get("_http_open_websockets", [])
            context["_http_open_websockets"].append(websocket)

            if access_log:
                http_logger.info(
                    '[{}] {} {} "OPEN {}{}" {} "{}" {}'.format(
                        RequestHandler.colorize_status("websocket", 101),
                        request_ip,
                        '"{}"'.format(request._cache["auth"].login.replace('"', ""))
                        if request._cache.get("auth") and getattr(request._cache.get("auth"), "login", None)
                        else "-",
                        request.path,
                        "?{}".format(request.query_string) if request.query_string else "",
                        request._cache.get("websocket_uuid", ""),
                        request.headers.get("User-Agent", "").replace('"', ""),
                        "-",
                    )
                )

            kwargs = dict(original_kwargs)
            if "(" in pattern:
                result = compiled_pattern.match(request.path)
                if result:
                    for k, v in result.groupdict().items():
                        kwargs[k] = v

            if len(values.args) - (len(values.defaults) if values.defaults else 0) >= 3:
                # If the function takes a third required argument the value will be filled with the request object
                a = a + (request,)
            if "request" in values.args and (
                len(values.args) - (len(values.defaults) if values.defaults else 0) < 3 or values.args[2] != "request"
            ):
                kwargs["request"] = request

            try:
                routine = func(*(obj, websocket, *a), **merge_dicts(kwargs, kw))
                callback_functions: Optional[Union[Tuple[Callable, Callable], Tuple[Callable], Callable]] = (
                    (await routine) if inspect.isawaitable(routine) else routine
                )
            except Exception as e:
                logging.getLogger("exception").exception("Uncaught exception: {}".format(str(e)))
                try:
                    await websocket.close()
                except Exception:
                    pass

                try:
                    context["_http_open_websockets"].remove(websocket)
                except Exception:
                    pass

                if access_log:
                    http_logger.info(
                        '[{}] {} {} "{} {}{}" {} "{}" {}'.format(
                            RequestHandler.colorize_status("websocket", 500),
                            request_ip,
                            '"{}"'.format(request._cache["auth"].login.replace('"', ""))
                            if request._cache.get("auth") and getattr(request._cache.get("auth"), "login", None)
                            else "-",
                            RequestHandler.colorize_status("ERROR", 500),
                            request.path,
                            "?{}".format(request.query_string) if request.query_string else "",
                            request._cache.get("websocket_uuid", ""),
                            request.headers.get("User-Agent", "").replace('"', ""),
                            "-",
                        )
                    )

                return

            _receive_func = None
            _close_func = None

            if callback_functions and isinstance(callback_functions, tuple):
                if len(callback_functions) == 2:
                    _receive_func, _close_func = cast(Tuple[Callable, Callable], callback_functions)
                elif len(callback_functions) == 1:
                    (_receive_func,) = cast(Tuple[Callable], callback_functions)
            elif callback_functions:
                _receive_func = callback_functions

            try:
                async for message in websocket:
                    if message.type == WSMsgType.TEXT:
                        if _receive_func:
                            try:
                                await _receive_func(message.data)
                            except Exception as e:
                                logging.getLogger("exception").exception("Uncaught exception: {}".format(str(e)))
                    elif message.type == WSMsgType.ERROR:
                        if not context.get("log_level") or context.get("log_level") in ["DEBUG"]:
                            ws_exception = websocket.exception()
                            if isinstance(ws_exception, (EofStream, RuntimeError)):
                                pass
                            elif isinstance(ws_exception, Exception):
                                logging.getLogger("exception").exception(
                                    "Uncaught exception: {}".format(str(ws_exception))
                                )
                            else:
                                http_logger.warning('Websocket exception: "{}"'.format(ws_exception))
                    elif message.type == WSMsgType.CLOSED:
                        break  # noqa
            except Exception:
                pass
            finally:
                if _close_func:
                    try:
                        await _close_func()
                    except Exception as e:
                        logging.getLogger("exception").exception("Uncaught exception: {}".format(str(e)))
                try:
                    await websocket.close()
                except Exception:
                    pass

                try:
                    context["_http_open_websockets"].remove(websocket)
                except Exception:
                    pass

        return await cls.request_handler(obj, context, _func, "GET", url, pre_handler_func=_pre_handler_func)

    @staticmethod
    async def start_server(obj: Any, context: Dict) -> Optional[Callable]:
        if context.get("_http_server_started"):
            return None
        context["_http_server_started"] = True

        http_options = context.get("options", {}).get("http", {})

        server_header = http_options.get("server_header", "tomodachi")
        access_log = http_options.get("access_log", True)

        logger_handler = None
        if isinstance(access_log, str):
            from logging.handlers import WatchedFileHandler  # noqa  # isort:skip

            try:
                wfh = WatchedFileHandler(filename=access_log)
            except FileNotFoundError as e:
                http_logger.warning('Unable to use file for access log - invalid path ("{}")'.format(access_log))
                raise HttpException(str(e)) from e
            except PermissionError as e:
                http_logger.warning('Unable to use file for access log - invalid permissions ("{}")'.format(access_log))
                raise HttpException(str(e)) from e
            wfh.setLevel(logging.DEBUG)
            http_logger.setLevel(logging.DEBUG)
            http_logger.info('Logging to "{}"'.format(access_log))
            logger_handler = wfh
            http_logger.addHandler(logger_handler)

        async def _start_server() -> None:
            loop = asyncio.get_event_loop()

            logging.getLogger("aiohttp.access").setLevel(logging.WARNING)

            async def request_handler_func(
                request: web.Request, handler: Callable
            ) -> Union[web.Response, web.FileResponse]:
                response: Union[web.Response, web.FileResponse]
                request_ip = RequestHandler.get_request_ip(request, context)

                if not request_ip:
                    # Transport broken before request handling started, ignore request
                    response = web.Response(status=499, headers={hdrs.SERVER: server_header or ""})
                    response._eof_sent = True
                    response.force_close()

                    return response

                if request.headers.get("Authorization"):
                    try:
                        request._cache["auth"] = BasicAuth.decode(request.headers.get("Authorization", ""))
                    except ValueError:
                        pass

                timer = time.time() if access_log else 0
                response = web.Response(status=503, headers={})
                try:
                    response = await handler(request)
                except web.HTTPException as e:
                    error_handler = context.get("_http_error_handler", {}).get(e.status, None)
                    if error_handler:
                        response = await error_handler(request)
                    else:
                        response = e
                        response.body = str(e).encode("utf-8")
                except Exception as e:
                    error_handler = context.get("_http_error_handler", {}).get(500, None)
                    logging.getLogger("exception").exception("Uncaught exception: {}".format(str(e)))
                    if error_handler:
                        response = await error_handler(request)
                    else:
                        response = web.HTTPInternalServerError()
                        response.body = b""
                finally:
                    if not request.transport:
                        response = web.Response(status=499, headers={})
                        response._eof_sent = True

                    request_version = (
                        (request.version.major, request.version.minor)
                        if isinstance(request.version, HttpVersion)
                        else (1, 0)
                    )

                    if access_log:
                        request_time = time.time() - timer
                        version_string = None
                        if isinstance(request.version, HttpVersion):
                            version_string = "HTTP/{}.{}".format(request_version[0], request_version[1])

                        if not request._cache.get("is_websocket"):
                            status_code = response.status if response is not None else 500
                            ignore_logging = getattr(handler, "ignore_logging", False)
                            if ignore_logging is True:
                                pass
                            elif isinstance(ignore_logging, (list, tuple)) and status_code in ignore_logging:
                                pass
                            else:
                                http_logger.info(
                                    '[{}] [{}] {} {} "{} {}{}{}" {} {} "{}" {}'.format(
                                        RequestHandler.colorize_status("http", status_code),
                                        RequestHandler.colorize_status(status_code),
                                        request_ip,
                                        '"{}"'.format(request._cache["auth"].login.replace('"', ""))
                                        if request._cache.get("auth")
                                        and getattr(request._cache.get("auth"), "login", None)
                                        else "-",
                                        request.method,
                                        request.path,
                                        "?{}".format(request.query_string) if request.query_string else "",
                                        " {}".format(version_string) if version_string else "",
                                        response.content_length
                                        if response is not None and response.content_length is not None
                                        else "-",
                                        request.content_length if request.content_length is not None else "-",
                                        request.headers.get("User-Agent", "").replace('"', ""),
                                        "{0:.5f}s".format(round(request_time, 5)),
                                    )
                                )
                        else:
                            http_logger.info(
                                '[{}] {} {} "CLOSE {}{}" {} "{}" {}'.format(
                                    RequestHandler.colorize_status("websocket", 101),
                                    request_ip,
                                    '"{}"'.format(request._cache["auth"].login.replace('"', ""))
                                    if request._cache.get("auth") and getattr(request._cache.get("auth"), "login", None)
                                    else "-",
                                    request.path,
                                    "?{}".format(request.query_string) if request.query_string else "",
                                    request._cache.get("websocket_uuid", ""),
                                    request.headers.get("User-Agent", "").replace('"', ""),
                                    "{0:.5f}s".format(round(request_time, 5)),
                                )
                            )

                    if response is not None:
                        response.headers[hdrs.SERVER] = server_header or ""

                        if request_version in ((1, 0), (1, 1)) and not request._cache.get("is_websocket"):
                            use_keepalive = False
                            if context["_http_tcp_keepalive"] and request.keep_alive and request.protocol:
                                use_keepalive = True
                                if any(
                                    [
                                        # keep-alive timeout not set or is non-positive
                                        (
                                            not context["_http_keepalive_timeout"]
                                            or context["_http_keepalive_timeout"] <= 0
                                        ),
                                        # keep-alive request count has passed configured max for this connection
                                        (
                                            context["_http_max_keepalive_requests"]
                                            and request.protocol._request_count
                                            >= context["_http_max_keepalive_requests"]
                                        ),
                                        # keep-alive time has passed configured max for this connection
                                        (
                                            context["_http_max_keepalive_time"]
                                            and time.time()
                                            > getattr(request.protocol, "_connection_start_time", 0)
                                            + context["_http_max_keepalive_time"]
                                        ),
                                    ]
                                ):
                                    use_keepalive = False

                            if use_keepalive:
                                response.headers[hdrs.CONNECTION] = "keep-alive"
                                response.headers[hdrs.KEEP_ALIVE] = "timeout={}{}".format(
                                    request.protocol._keepalive_timeout,
                                    ", max={}".format(context["_http_max_keepalive_requests"])
                                    if context["_http_max_keepalive_requests"]
                                    else "",
                                )
                            else:
                                response.headers[hdrs.CONNECTION] = "close"
                                response.force_close()

                        if not context["_http_tcp_keepalive"] and not request._cache.get("is_websocket"):
                            response.force_close()

                    if isinstance(response, (web.HTTPException, web.HTTPInternalServerError)):
                        raise response

                    return response

            @web.middleware
            async def middleware(request: web.Request, handler: Callable) -> Union[web.Response, web.FileResponse]:
                increase_execution_context_value("http_current_tasks")
                increase_execution_context_value("http_total_tasks")
                task = asyncio.ensure_future(request_handler_func(request, handler))
                context["_http_active_requests"] = context.get("_http_active_requests", set())
                context["_http_active_requests"].add(task)
                try:
                    await asyncio.shield(task)
                except asyncio.CancelledError:
                    try:
                        await task
                        decrease_execution_context_value("http_current_tasks")
                        try:
                            context["_http_active_requests"].remove(task)
                        except KeyError:
                            pass
                        return task.result()
                    except Exception:
                        decrease_execution_context_value("http_current_tasks")
                        try:
                            context["_http_active_requests"].remove(task)
                        except KeyError:
                            pass
                        raise
                except Exception:
                    decrease_execution_context_value("http_current_tasks")
                    try:
                        context["_http_active_requests"].remove(task)
                    except KeyError:
                        pass
                    raise
                except BaseException:
                    decrease_execution_context_value("http_current_tasks")
                    try:
                        context["_http_active_requests"].remove(task)
                    except KeyError:
                        pass
                    raise
                decrease_execution_context_value("http_current_tasks")
                try:
                    context["_http_active_requests"].remove(task)
                except KeyError:
                    pass

                return task.result()

            client_max_size_option = (
                http_options.get("client_max_size")
                or http_options.get("max_buffer_size")
                or http_options.get("max_upload_size")
                or "100M"
            )
            client_max_size_option_str = str(client_max_size_option).upper()
            client_max_size = (1024**2) * 100
            try:
                if (
                    client_max_size_option
                    and isinstance(client_max_size_option, str)
                    and (client_max_size_option_str.endswith("G") or client_max_size_option_str.endswith("GB"))
                ):
                    client_max_size = int(
                        re.sub(cast(str, r"^([0-9]+)GB?$"), cast(str, r"\1"), client_max_size_option_str)
                    ) * (1024**3)
                elif (
                    client_max_size_option
                    and isinstance(client_max_size_option, str)
                    and (client_max_size_option_str.endswith("M") or client_max_size_option_str.endswith("MB"))
                ):
                    client_max_size = int(re.sub(r"^([0-9]+)MB?$", r"\1", client_max_size_option_str)) * (1024**2)
                elif (
                    client_max_size_option
                    and isinstance(client_max_size_option, str)
                    and (client_max_size_option_str.endswith("K") or client_max_size_option_str.endswith("KB"))
                ):
                    client_max_size = int(re.sub(r"^([0-9]+)KB?$", r"\1", client_max_size_option_str)) * 1024
                elif (
                    client_max_size_option
                    and isinstance(client_max_size_option, str)
                    and (client_max_size_option_str.endswith("B"))
                ):
                    client_max_size = int(re.sub(r"^([0-9]+)B?$", r"\1", client_max_size_option_str))
                elif client_max_size_option:
                    client_max_size = int(client_max_size_option)
            except Exception:
                raise ValueError(
                    "Bad value for http option client_max_size: {}".format(str(client_max_size_option))
                ) from None
            if client_max_size >= 0 and client_max_size < 1024:
                raise ValueError(
                    "Too low value for http option client_max_size: {} ({})".format(
                        str(client_max_size_option), client_max_size_option
                    )
                )
            if client_max_size > 1024**3:
                raise ValueError(
                    "Too high value for http option client_max_size: {} ({})".format(
                        str(client_max_size_option), client_max_size_option
                    )
                )

            app: web.Application = web.Application(middlewares=[middleware], client_max_size=client_max_size)
            app._set_loop(None)
            for method, pattern, handler, route_context in context.get("_http_routes", []):
                try:
                    compiled_pattern = re.compile(pattern)
                except re.error as exc:
                    raise ValueError("Bad http route pattern '{}': {}".format(pattern, exc)) from None
                ignore_logging = route_context.get("ignore_logging", False)
                setattr(handler, "ignore_logging", ignore_logging)
                resource = DynamicResource(compiled_pattern)
                app.router.register_resource(resource)
                if method.upper() == "GET":
                    resource.add_route("HEAD", handler, expect_handler=None)
                resource.add_route(method.upper(), handler, expect_handler=None)

            context["_http_accept_new_requests"] = True

            port = http_options.get("port", 9700)
            host = http_options.get("host", "0.0.0.0")
            if port is True:
                raise ValueError("Bad value for http option port: {}".format(str(port)))

            # Configuration settings for keep-alive could use some refactoring

            keepalive_timeout_option = http_options.get("keepalive_timeout", 0) or http_options.get(
                "keepalive_expiry", 0
            )
            keepalive_timeout = 0
            if keepalive_timeout_option is None or keepalive_timeout_option is False:
                keepalive_timeout_option = 0
            if keepalive_timeout_option is True:
                raise ValueError(
                    "Bad value for http option keepalive_timeout: {}".format(str(keepalive_timeout_option))
                )
            try:
                keepalive_timeout = int(keepalive_timeout_option) if keepalive_timeout_option is not None else 0
            except Exception:
                raise ValueError(
                    "Bad value for http option keepalive_timeout: {}".format(str(keepalive_timeout_option))
                ) from None

            tcp_keepalive = False
            if keepalive_timeout > 0:
                tcp_keepalive = True
            else:
                tcp_keepalive = False
                keepalive_timeout = 0

            max_keepalive_requests_option = http_options.get("max_keepalive_requests", None)
            max_keepalive_requests = None
            if max_keepalive_requests_option is None or max_keepalive_requests_option is False:
                max_keepalive_requests_option = None
            if max_keepalive_requests_option is True:
                raise ValueError(
                    "Bad value for http option max_keepalive_requests: {}".format(str(max_keepalive_requests_option))
                )
            try:
                if max_keepalive_requests_option is not None:
                    max_keepalive_requests = int(max_keepalive_requests_option)
                if max_keepalive_requests == 0:
                    max_keepalive_requests = None
            except Exception:
                raise ValueError(
                    "Bad value for http option max_keepalive_requests: {}".format(str(max_keepalive_requests_option))
                ) from None
            if not tcp_keepalive and max_keepalive_requests:
                raise ValueError(
                    "HTTP keep-alive must be enabled to use http option max_keepalive_requests - a http.keepalive_timeout option value is required"
                ) from None

            max_keepalive_time_option = http_options.get("max_keepalive_time", None)
            max_keepalive_time = None
            if max_keepalive_time_option is None or max_keepalive_time_option is False:
                max_keepalive_time_option = None
            if max_keepalive_time_option is True:
                raise ValueError(
                    "Bad value for http option max_keepalive_time: {}".format(str(max_keepalive_time_option))
                )
            try:
                if max_keepalive_time_option is not None:
                    max_keepalive_time = int(max_keepalive_time_option)
                if max_keepalive_time == 0:
                    max_keepalive_time = None
            except Exception:
                raise ValueError(
                    "Bad value for http option max_keepalive_time: {}".format(str(max_keepalive_time_option))
                ) from None
            if not tcp_keepalive and max_keepalive_time:
                raise ValueError(
                    "HTTP keep-alive must be enabled to use http option max_keepalive_time - a http.keepalive_timeout option value is required"
                ) from None

            reuse_port_default = True if platform.system() == "Linux" else False
            reuse_port = True if http_options.get("reuse_port", reuse_port_default) else False
            if reuse_port and platform.system() != "Linux":
                http_logger.warning(
                    "The http option reuse_port (socket.SO_REUSEPORT) can only enabled on Linux platforms - current "
                    f"platform is {platform.system()} - will revert option setting to not reuse ports"
                )
                reuse_port = False

            context["_http_tcp_keepalive"] = tcp_keepalive
            context["_http_keepalive_timeout"] = keepalive_timeout
            context["_http_max_keepalive_requests"] = max_keepalive_requests
            context["_http_max_keepalive_time"] = max_keepalive_time

            set_execution_context(
                {
                    "http_enabled": True,
                    "http_current_tasks": 0,
                    "http_total_tasks": 0,
                    "aiohttp_version": aiohttp_version,
                }
            )

            try:
                app.freeze()
                web_server = Server(
                    app._handle,
                    request_factory=app._make_request,
                    server_header=server_header or "",
                    access_log=access_log,
                    keepalive_timeout=keepalive_timeout,
                    tcp_keepalive=tcp_keepalive,
                )
                if reuse_port:
                    if not port:
                        http_logger.warning(
                            "The http option reuse_port (socket option SO_REUSEPORT) is enabled by default on Linux - "
                            "listening on random ports with SO_REUSEPORT is dangerous - please double check your intent"
                        )
                    elif str(port) in HttpTransport.server_port_mapping.values():
                        http_logger.warning(
                            "The http option reuse_port (socket option SO_REUSEPORT) is enabled by default on Linux - "
                            "different service classes should not use the same port ({})".format(port)
                        )
                if port:
                    HttpTransport.server_port_mapping[web_server] = str(port)
                server_task = loop.create_server(web_server, host, port, reuse_port=reuse_port)
                server = await server_task
            except OSError as e:
                context["_http_accept_new_requests"] = False
                error_message = re.sub(".*: ", "", e.strerror)
                http_logger.warning(
                    "Unable to bind service [http] to http://{}:{}/ ({})".format(
                        "127.0.0.1" if host == "0.0.0.0" else host, port, error_message
                    )
                )
                raise HttpException(str(e), log_level=context.get("log_level")) from e

            if server.sockets:
                socket_address = server.sockets[0].getsockname()
                if socket_address:
                    port = int(socket_address[1])
                    HttpTransport.server_port_mapping[web_server] = str(port)
            context["_http_port"] = port

            stop_method = getattr(obj, "_stop_service", None)

            async def stop_service(*args: Any, **kwargs: Any) -> None:
                context["_http_tcp_keepalive"] = False

                server.close()
                await server.wait_closed()

                HttpTransport.server_port_mapping.pop(web_server, None)

                shutdown_sleep = 0
                if len(web_server.connections):
                    shutdown_sleep = 1
                    await asyncio.sleep(1)

                if not tcp_keepalive:
                    context["_http_accept_new_requests"] = False

                open_websockets = context.get("_http_open_websockets", [])[:]
                if open_websockets:
                    http_logger.info("Closing {} websocket connection(s)".format(len(open_websockets)))
                    tasks = []
                    for websocket in open_websockets:
                        try:
                            tasks.append(asyncio.ensure_future(websocket.close()))
                        except Exception:
                            pass
                    try:
                        await asyncio.wait_for(
                            asyncio.shield(asyncio.gather(*tasks, return_exceptions=True)), timeout=10
                        )
                        await asyncio.sleep(1)
                    except (Exception, asyncio.TimeoutError, asyncio.CancelledError):
                        pass
                    context["_http_open_websockets"] = []

                termination_grace_period_seconds = 30
                try:
                    termination_grace_period_seconds = int(
                        http_options.get("termination_grace_period_seconds") or termination_grace_period_seconds
                    )
                except Exception:
                    pass

                log_wait_message = True if termination_grace_period_seconds >= 2 else False

                if tcp_keepalive and len(web_server.connections):
                    wait_start_time = time.time()

                    while wait_start_time + max(2, termination_grace_period_seconds) > time.time():
                        active_requests = context.get("_http_active_requests", set())
                        if not active_requests and not len(web_server.connections):
                            break

                        if log_wait_message:
                            log_wait_message = False
                            if len(web_server.connections) and len(web_server.connections) != len(active_requests):
                                http_logger.info(
                                    "Waiting for {} keep-alive connection(s) to close".format(
                                        len(web_server.connections)
                                    )
                                )
                            if active_requests:
                                http_logger.info(
                                    "Waiting for {} active request(s) to complete - grace period of {} seconds".format(
                                        len(active_requests), termination_grace_period_seconds
                                    )
                                )

                        await asyncio.sleep(0.25)

                    termination_grace_period_seconds -= int(time.time() - wait_start_time)

                context["_http_accept_new_requests"] = False

                active_requests = context.get("_http_active_requests", set())
                if active_requests:
                    if log_wait_message:
                        http_logger.info(
                            "Waiting for {} active request(s) to complete - grace period of {} seconds".format(
                                len(active_requests), termination_grace_period_seconds
                            )
                        )

                    try:
                        await asyncio.wait_for(
                            asyncio.shield(asyncio.gather(*active_requests, return_exceptions=True)),
                            timeout=max(2, termination_grace_period_seconds),
                        )
                        await asyncio.sleep(1)
                        active_requests = context.get("_http_active_requests", set())
                    except (Exception, asyncio.TimeoutError, asyncio.CancelledError):
                        active_requests = context.get("_http_active_requests", set())
                        if active_requests:
                            http_logger.warning(
                                "All requests did not gracefully finish execution - {} request(s) remaining".format(
                                    len(active_requests)
                                )
                            )
                    context["_http_active_requests"] = set()

                if shutdown_sleep > 0:
                    await asyncio.sleep(shutdown_sleep)

                if len(web_server.connections):
                    http_logger.warning(
                        "The remaining {} open TCP connections will be forcefully closed".format(
                            len(web_server.connections)
                        )
                    )
                    await app.shutdown()
                    await asyncio.sleep(1)
                else:
                    await app.shutdown()

                if logger_handler:
                    http_logger.removeHandler(logger_handler)
                await app.cleanup()
                if stop_method:
                    await stop_method(*args, **kwargs)

            setattr(obj, "_stop_service", stop_service)

            for method, pattern, handler, route_context in context.get("_http_routes", []):
                for registry in getattr(obj, "discovery", []):
                    if getattr(registry, "add_http_endpoint", None):
                        await registry.add_http_endpoint(obj, host, port, method, pattern)

            http_logger.info(
                "Listening [http] on http://{}:{}/".format("127.0.0.1" if host == "0.0.0.0" else host, port)
            )

        return _start_server


async def resolve_response(
    value: Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response],
    request: Optional[web.Request] = None,
    context: Dict = None,
    status_code: Optional[Union[str, int]] = None,
    default_content_type: Optional[str] = None,
    default_charset: Optional[str] = None,
) -> Union[web.Response, web.FileResponse]:
    return resolve_response_sync(
        value=value,
        request=request,
        context=context,
        status_code=status_code,
        default_content_type=default_content_type,
        default_charset=default_charset,
    )


def resolve_response_sync(
    value: Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response],
    request: Optional[web.Request] = None,
    context: Dict = None,
    status_code: Optional[Union[str, int]] = None,
    default_content_type: Optional[str] = None,
    default_charset: Optional[str] = None,
) -> Union[web.Response, web.FileResponse]:
    if not context:
        context = {}
    if isinstance(value, Response):
        return value.get_aiohttp_response(
            context, default_content_type=default_content_type, default_charset=default_charset
        )
    if isinstance(value, web.FileResponse):
        return value

    status = (
        int(status_code)
        if status_code
        else (request is not None and request._cache.get("error_status_code", 200)) or 200
    )
    headers = None
    if isinstance(value, dict):
        body = value.get("body")
        _status: Optional[SupportsInt] = value.get("status")
        if _status and isinstance(_status, (int, str, bytes)):
            status = int(_status)
        _returned_headers = value.get("headers")
        if _returned_headers:
            returned_headers: Union[Mapping[str, Any], Iterable[Tuple[str, Any]]] = _returned_headers
            headers = CIMultiDict(returned_headers)
    elif isinstance(value, list) or isinstance(value, tuple):
        _status = value[0]
        if _status and isinstance(_status, (int, str, bytes)):
            status = int(_status)
        body = value[1]
        if len(value) > 2:
            returned_headers = value[2]
            headers = CIMultiDict(returned_headers)
    elif isinstance(value, web.Response):
        return value
    else:
        if value is None:
            value = ""
        body = value

    return Response(
        body=body, status=status, headers=headers, content_type=default_content_type, charset=default_charset
    ).get_aiohttp_response(context)


async def get_http_response_status(
    value: Union[str, bytes, Dict, List, Tuple, web.Response, web.FileResponse, Response, Exception],
    request: Optional[web.Request] = None,
    verify_transport: bool = True,
) -> Optional[int]:
    if isinstance(value, Exception) or isinstance(value, web.HTTPException):
        status_code = int(getattr(value, "status", 500)) if value is not None else 500
        return status_code
    else:
        response = resolve_response_sync(value, request=request)
        status_code = int(response.status) if response is not None else 500
        if verify_transport and request is not None and request.transport is None:
            return 499
        else:
            return status_code


def get_http_response_status_sync(
    value: Any, request: Optional[web.Request] = None, verify_transport: bool = True
) -> Optional[int]:
    if isinstance(value, Exception) or isinstance(value, web.HTTPException):
        status_code = int(getattr(value, "status", 500)) if value is not None else 500
        return status_code

    if verify_transport and request is not None and hasattr(request, "transport") and request.transport is None:
        return 499

    if isinstance(value, Response) and value._status:
        return int(value._status)
    elif isinstance(value, (web.Response, web.FileResponse)) and value.status:
        return int(value.status)
    elif isinstance(value, dict):
        _status: Optional[SupportsInt] = value.get("status")
        if _status and isinstance(_status, (int, str, bytes)):
            return int(_status)
    elif isinstance(value, list) or isinstance(value, tuple):
        _status = value[0]
        if _status and isinstance(_status, (int, str, bytes)):
            return int(_status)
    elif value and hasattr(value, "_status") and getattr(value, "_status", None):
        return int(getattr(value, "_status"))
    elif value and hasattr(value, "status") and getattr(value, "status", None):
        return int(getattr(value, "status"))

    return int((request is not None and request._cache.get("error_status_code", 200)) or 200)


__http = HttpTransport.decorator(HttpTransport.request_handler)
__http_error = HttpTransport.decorator(HttpTransport.error_handler)
__http_static = HttpTransport.decorator(HttpTransport.static_request_handler)

__websocket = HttpTransport.decorator(HttpTransport.websocket_handler)
__ws = HttpTransport.decorator(HttpTransport.websocket_handler)


def http(
    method: Union[str, List[str], Tuple[str, ...]],
    url: str,
    *,
    ignore_logging: Union[bool, List[int], Tuple[int, ...]] = False,
    pre_handler_func: Optional[Callable] = None,
) -> Callable:
    return cast(Callable, __http(method, url, ignore_logging=ignore_logging, pre_handler_func=pre_handler_func))


def http_error(status_code: int) -> Callable:
    return cast(Callable, __http_error(status_code))


def http_static(
    path: str, base_url: str, *, ignore_logging: Union[bool, List[int], Tuple[int, ...]] = False
) -> Callable:
    return cast(Callable, __http_static(path, base_url, ignore_logging=ignore_logging))


def websocket(url: str) -> Callable:
    return cast(Callable, __websocket(url))


def ws(url: str) -> Callable:
    return cast(Callable, __ws(url))



import asyncio
import datetime
import importlib
import logging
import os
import platform
import signal
import sys
import time
from typing import Any, Dict, List, Optional, Set, Union, cast

import tomodachi
import tomodachi.__version__
import tomodachi.container
import tomodachi.importer
import tomodachi.invoker
from tomodachi.container import ServiceContainer
from tomodachi.helpers.execution_context import clear_execution_context, clear_services, set_execution_context
from tomodachi.helpers.safe_modules import SAFE_MODULES
from tomodachi.importer import ServiceImporter

CancelledError = asyncio.CancelledError
try:
    asyncioexceptions = getattr(asyncio, "exceptions")
    if asyncioexceptions:
        _CancelledError = asyncioexceptions.CancelledError
except (Exception, ModuleNotFoundError, ImportError):
    _CancelledError = asyncio.CancelledError


class ServiceLauncher(object):
    _close_waiter: Optional[asyncio.Future] = None
    _stopped_waiter: Optional[asyncio.Future] = None
    restart_services = False
    services: Set = set()

    @classmethod
    def stop_services(cls) -> None:
        asyncio.ensure_future(cls._stop_services())

    @classmethod
    async def _stop_services(cls) -> None:
        if cls._close_waiter and not cls._close_waiter.done():
            cls._close_waiter.set_result(None)
            for service in cls.services:
                try:
                    service.stop_service()
                except Exception:
                    pass
            if cls._stopped_waiter:
                cls._stopped_waiter.set_result(None)
        if cls._stopped_waiter:
            await cls._stopped_waiter

    @classmethod
    def run_until_complete(
        cls,
        service_files: Union[List, set],
        configuration: Optional[Dict] = None,
        watcher: Any = None,
    ) -> None:
        def sigintHandler(*args: Any) -> None:
            sys.stdout.write("\b\b\r")
            sys.stdout.flush()
            logging.getLogger("system").warning("Received <ctrl+c> interrupt [SIGINT]")
            cls.restart_services = False

        def sigtermHandler(*args: Any) -> None:
            logging.getLogger("system").warning("Received termination signal [SIGTERM]")
            cls.restart_services = False

        logging.basicConfig(level=logging.DEBUG)

        loop = asyncio.get_event_loop()
        if loop and loop.is_closed():
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        for signame in ("SIGINT", "SIGTERM"):
            loop.add_signal_handler(getattr(signal, signame), cls.stop_services)

        signal.siginterrupt(signal.SIGTERM, False)
        signal.siginterrupt(signal.SIGUSR1, False)
        signal.signal(signal.SIGINT, sigintHandler)
        signal.signal(signal.SIGTERM, sigtermHandler)

        watcher_future = None
        if watcher:

            async def _watcher_restart(updated_files: Union[List, set]) -> None:
                cls.restart_services = True

                for file in service_files:
                    try:
                        ServiceImporter.import_service_file(file)
                    except (SyntaxError, IndentationError) as e:
                        logging.getLogger("exception").exception("Uncaught exception: {}".format(str(e)))
                        logging.getLogger("watcher.restart").warning("Service cannot restart due to errors")
                        cls.restart_services = False
                        return

                pre_import_current_modules = [m for m in sys.modules.keys()]
                cwd = os.getcwd()
                for file in updated_files:
                    if file.lower().endswith(".py"):
                        module_name = file[:-3].replace("/", ".")
                        module_name_full_path = "{}/{}".format(os.path.realpath(cwd), file)[:-3].replace("/", ".")
                        try:
                            for m in pre_import_current_modules:
                                if m == module_name or (len(m) > len(file) and module_name_full_path.endswith(m)):
                                    ServiceImporter.import_module(file)
                        except (SyntaxError, IndentationError) as e:
                            logging.getLogger("exception").exception("Uncaught exception: {}".format(str(e)))
                            logging.getLogger("watcher.restart").warning("Service cannot restart due to errors")
                            cls.restart_services = False
                            return

                logging.getLogger("watcher.restart").warning("Restarting services")
                cls.stop_services()

            watcher_future = loop.run_until_complete(watcher.watch(loop=loop, callback_func=_watcher_restart))

        cls.restart_services = True
        init_modules = [m for m in sys.modules.keys()]

        restarting = False
        while cls.restart_services:
            init_timestamp = time.time()
            init_timestamp_str = datetime.datetime.utcfromtimestamp(init_timestamp).isoformat() + "Z"

            process_id = os.getpid()

            event_loop_alias = ""
            event_loop_version = ""
            try:
                if "uvloop." in str(loop.__class__):
                    event_loop_alias = "uvloop"
                    import uvloop  # noqa  # isort:skip

                    event_loop_version = str(uvloop.__version__)
                elif "asyncio." in str(loop.__class__):
                    event_loop_alias = "asyncio"
                else:
                    event_loop_alias = "{}.{}".format(loop.__class__.__module__, loop.__class__.__name__)
            except Exception:
                event_loop_alias = str(loop)

            clear_services()
            clear_execution_context()
            set_execution_context(
                {
                    "tomodachi_version": tomodachi.__version__,
                    "python_version": platform.python_version(),
                    "system_platform": platform.system(),
                    "process_id": process_id,
                    "init_timestamp": init_timestamp_str,
                    "event_loop": event_loop_alias,
                }
            )

            if event_loop_alias == "uvloop" and event_loop_version:
                set_execution_context(
                    {
                        "uvloop_version": event_loop_version,
                    }
                )

            if watcher:
                tz: Any = None
                utc_tz: Any = None

                try:
                    import pytz  # noqa  # isort:skip
                    import tzlocal  # noqa  # isort:skip

                    utc_tz = pytz.UTC
                    try:
                        tz = tzlocal.get_localzone()
                        if not tz:
                            tz = pytz.UTC
                    except Exception:
                        tz = pytz.UTC
                except Exception:
                    pass

                init_local_datetime = (
                    datetime.datetime.fromtimestamp(init_timestamp)
                    if tz and tz is not utc_tz and str(tz) != "UTC"
                    else datetime.datetime.utcfromtimestamp(init_timestamp)
                )

                print("---")
                print("Starting tomodachi services (pid: {}) ...".format(process_id))
                for file in service_files:
                    print("* {}".format(file))

                print()
                print(
                    "Current version: tomodachi {} on Python {}".format(
                        tomodachi.__version__, platform.python_version()
                    )
                )
                print(
                    "Event loop implementation: {}{}".format(
                        event_loop_alias, " {}".format(event_loop_version) if event_loop_version else ""
                    )
                )
                if tz:
                    print("Local time: {} {}".format(init_local_datetime.strftime("%B %d, %Y - %H:%M:%S,%f"), str(tz)))
                print("Timestamp in UTC: {}".format(init_timestamp_str))
                print()
                print("File watcher is active - code changes will automatically restart services")
                print("Quit running services with <ctrl+c>")
                print()

            tomodachi.SERVICE_EXIT_CODE = tomodachi.DEFAULT_SERVICE_EXIT_CODE

            cls._close_waiter = asyncio.Future()
            cls._stopped_waiter = asyncio.Future()
            cls.restart_services = False

            try:
                cls.services = set(
                    [
                        ServiceContainer(ServiceImporter.import_service_file(file), configuration)
                        for file in service_files
                    ]
                )
                result = loop.run_until_complete(
                    asyncio.wait([asyncio.ensure_future(service.run_until_complete()) for service in cls.services])
                )
                exception = [v.exception() for v in [value for value in result if value][0] if v.exception()]
                if exception:
                    raise cast(Exception, exception[0])
            except tomodachi.importer.ServicePackageError:
                pass
            except Exception as e:
                logging.getLogger("exception").exception("Uncaught exception: {}".format(str(e)))

                if isinstance(e, ModuleNotFoundError):  # pragma: no cover
                    missing_module_name = str(getattr(e, "name", None) or "")
                    if missing_module_name:
                        color = ""
                        color_reset = ""
                        try:
                            import colorama  # noqa  # isort:skip

                            color = colorama.Fore.WHITE + colorama.Back.RED
                            color_reset = colorama.Style.RESET_ALL
                        except Exception:
                            pass

                        print("")
                        print(
                            "{}[fatal error] The '{}' package is missing or cannot be imported.{}".format(
                                color, missing_module_name, color_reset
                            )
                        )
                        print("")

                if restarting:
                    logging.getLogger("watcher.restart").warning("Service cannot restart due to errors")
                    logging.getLogger("watcher.restart").warning("Trying again in 1.5 seconds")
                    loop.run_until_complete(asyncio.wait([asyncio.sleep(1.5)]))
                    if cls._close_waiter and not cls._close_waiter.done():
                        cls.restart_services = True
                    else:
                        for signame in ("SIGINT", "SIGTERM"):
                            loop.remove_signal_handler(getattr(signal, signame))
                else:
                    for signame in ("SIGINT", "SIGTERM"):
                        loop.remove_signal_handler(getattr(signal, signame))

            current_modules = [m for m in sys.modules.keys()]
            for m in current_modules:
                if m not in init_modules and m not in SAFE_MODULES:
                    del sys.modules[m]

            importlib.reload(tomodachi.container)
            importlib.reload(tomodachi.invoker)
            importlib.reload(tomodachi.invoker.base)
            importlib.reload(tomodachi.importer)

            restarting = True

        if watcher:
            if watcher_future and not watcher_future.done():
                try:
                    watcher_future.set_result(None)
                except RuntimeError:  # pragma: no cover
                    watcher_future.cancel()
                if not watcher_future.done():  # pragma: no cover
                    try:
                        loop.run_until_complete(watcher_future)
                    except (Exception, CancelledError, _CancelledError):
                        pass



import asyncio
import datetime
import inspect
import logging
import random
import time
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

import pytz
import tzlocal

from tomodachi.helpers.crontab import get_next_datetime
from tomodachi.helpers.execution_context import (
    decrease_execution_context_value,
    increase_execution_context_value,
    set_execution_context,
)
from tomodachi.invoker import Invoker


class Scheduler(Invoker):
    close_waiter = None

    @classmethod
    async def schedule_handler(
        cls,
        obj: Any,
        context: Dict,
        func: Any,
        interval: Optional[Union[str, int]] = None,
        timestamp: Optional[str] = None,
        timezone: Optional[str] = None,
        immediately: Optional[bool] = False,
    ) -> Any:
        values = inspect.getfullargspec(func)
        original_kwargs = (
            {k: values.defaults[i] for i, k in enumerate(values.args[len(values.args) - len(values.defaults) :])}
            if values.defaults
            else {}
        )

        async def handler() -> None:
            increase_execution_context_value("scheduled_functions_current_tasks")
            try:
                kwargs = dict(original_kwargs)

                increase_execution_context_value("scheduled_functions_total_tasks")

                routine = func(*(obj,), **kwargs)
                if inspect.isawaitable(routine):
                    await routine

            except (Exception, asyncio.CancelledError) as e:
                logging.getLogger("exception").exception("Uncaught exception: {}".format(str(e)))
            except BaseException as e:
                logging.getLogger("exception").exception("Uncaught exception: {}".format(str(e)))
            finally:
                decrease_execution_context_value("scheduled_functions_current_tasks")

        context["_schedule_scheduled_functions"] = context.get("_schedule_scheduled_functions", [])
        context["_schedule_scheduled_functions"].append((interval, timestamp, timezone, immediately, func, handler))

        start_func = cls.start_scheduler(obj, context)
        return (await start_func) if start_func else None

    @classmethod
    def schedule_handler_with_interval(cls, interval: Union[str, int]) -> Callable:
        def _func(_: Any, obj: Any, context: Dict, func: Any) -> Any:
            return cls.schedule_handler(obj, context, func, interval=interval)

        return _func

    @staticmethod
    def next_call_at(
        current_time: float,
        interval: Optional[Union[str, int]] = None,
        timestamp: Optional[str] = None,
        timezone: Optional[str] = None,
    ) -> int:
        if not timezone:
            tz = tzlocal.get_localzone()
        else:
            try:
                tz = pytz.timezone(timezone or "")
            except Exception as e:
                raise Exception("Unknown timezone: {}".format(timezone)) from e
        local_tz = tzlocal.get_localzone()

        if interval is None and timestamp is not None:
            if isinstance(timestamp, str):
                try:
                    datetime_object = datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")
                    interval = "{} {} {} {} * {}".format(
                        datetime_object.minute,
                        datetime_object.hour,
                        datetime_object.day,
                        datetime_object.month,
                        datetime_object.year,
                    )
                    second_modifier = 1
                    if datetime_object.replace(tzinfo=tz) > datetime.datetime.fromtimestamp(current_time).replace(
                        tzinfo=local_tz
                    ):
                        second_modifier = -60
                    next_at = get_next_datetime(
                        interval,
                        datetime.datetime.fromtimestamp(current_time + second_modifier)
                        .replace(tzinfo=local_tz)
                        .astimezone(tz),
                    )
                    if not next_at:
                        return int(current_time + 60 * 60 * 24 * 365 * 100)
                    return int(next_at.timestamp() + datetime_object.second)
                except ValueError:
                    pass

                try:
                    datetime_object = datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M")
                    interval = "{} {} {} {} * {}".format(
                        datetime_object.minute,
                        datetime_object.hour,
                        datetime_object.day,
                        datetime_object.month,
                        datetime_object.year,
                    )
                    next_at = get_next_datetime(
                        interval,
                        datetime.datetime.fromtimestamp(current_time + 1).replace(tzinfo=local_tz).astimezone(tz),
                    )
                    if not next_at:
                        return int(current_time + 60 * 60 * 24 * 365 * 100)
                    return int(next_at.timestamp())
                except ValueError:
                    pass

                try:
                    datetime_object = datetime.datetime.strptime(timestamp, "%H:%M:%S")
                    datetime_object = datetime.datetime(
                        datetime.datetime.fromtimestamp(current_time).year,
                        datetime.datetime.fromtimestamp(current_time).month,
                        datetime.datetime.fromtimestamp(current_time).day,
                        datetime_object.hour,
                        datetime_object.minute,
                        datetime_object.second,
                    )
                    interval = "{} {} * * *".format(datetime_object.minute, datetime_object.hour)
                    second_modifier = 1

                    if datetime_object.replace(tzinfo=tz) > datetime.datetime.fromtimestamp(current_time).replace(
                        tzinfo=local_tz
                    ):
                        second_modifier = -60
                    next_at = get_next_datetime(
                        interval,
                        datetime.datetime.fromtimestamp(current_time + second_modifier)
                        .replace(tzinfo=local_tz)
                        .astimezone(tz),
                    )
                    if not next_at:
                        return int(current_time + 60 * 60 * 24 * 365 * 100)
                    return int(next_at.timestamp() + datetime_object.second)
                except ValueError:
                    pass

                try:
                    datetime_object = datetime.datetime.strptime(timestamp, "%H:%M")
                    interval = "{} {} * * *".format(datetime_object.minute, datetime_object.hour)
                    next_at = get_next_datetime(
                        interval,
                        datetime.datetime.fromtimestamp(current_time + 1).replace(tzinfo=local_tz).astimezone(tz),
                    )
                    if not next_at:
                        return int(current_time + 60 * 60 * 24 * 365 * 100)
                    return int(next_at.timestamp())
                except ValueError:
                    pass

                raise Exception("Invalid timestamp")

        if interval is not None:
            if isinstance(interval, int):
                return int(current_time + interval)

            interval_aliases: Dict[Tuple[str, ...], Union[str, int]] = {
                ("every second", "1s", "1 s", "1second", "1 second", "second", "secondly", "once per second"): 1,
                (
                    "every minute",
                    "1m",
                    "1 m",
                    "1minute",
                    "1 minute",
                    "minute",
                    "minutely",
                    "once per minute",
                ): "@minutely",
                ("every hour", "1h", "1 h", "1hour", "1 hour", "hour", "hourly", "once per hour"): "@hourly",
                ("every day", "1d", "1 d", "1day", "1 day", "day", "daily", "once per day", "nightly"): "@daily",
                ("every month", "1month", "1 month", "month", "monthly", "once per month"): "@monthly",
                (
                    "every year",
                    "1y",
                    "1 y",
                    "1year",
                    "1 year",
                    "year",
                    "yearly",
                    "once per year",
                    "annually",
                ): "@yearly",
                (
                    "monday",
                    "mondays",
                    "mon",
                    "every monday",
                    "once per monday",
                    "weekly",
                    "once per week",
                    "week",
                    "every week",
                ): "0 0 * * 1",
                ("tuesday", "tuesdays", "tue", "every tuesday", "once per tuesday"): "0 0 * * 2",
                ("wednesday", "wednesdays", "wed", "every wednesday", "once per wednesday"): "0 0 * * 3",
                ("thursday", "thursdays", "thu", "every thursday", "once per thursday"): "0 0 * * 4",
                ("friday", "fridays", "fri", "every friday", "once per friday"): "0 0 * * 5",
                ("saturday", "saturdays", "sat", "every saturday", "once per saturday"): "0 0 * * 6",
                ("sunday", "sundays", "sun", "every sunday", "once per sunday"): "0 0 * * 0",
                ("weekday", "weekdays", "every weekday"): "0 0 * * 1-5",
                ("weekend", "weekends", "every weekend"): "0 0 * * 0,6",
            }
            interval = interval.lower()

            if interval.endswith("s") or interval.endswith("seconds"):
                try:
                    interval = int(interval.replace("seconds", "").replace("s", "").replace(" ", ""))
                except ValueError:
                    pass

            try:
                interval_value: Union[str, int] = [v for k, v in interval_aliases.items() if interval in k][0]
            except IndexError:
                interval_value = interval
            if isinstance(interval_value, int):
                return int(current_time + interval_value)

            try:
                next_at = get_next_datetime(
                    interval_value,
                    datetime.datetime.fromtimestamp(current_time + 1).replace(tzinfo=local_tz).astimezone(tz),
                )
                if not next_at:
                    return int(current_time + 60 * 60 * 24 * 365 * 100)
                return int(next_at.timestamp())
            except Exception:
                raise Exception("Invalid interval")

        return int(current_time + 60 * 60 * 24 * 365 * 100)

    @staticmethod
    def get_timezone(timezone: Optional[str] = None) -> Optional[str]:
        if timezone:
            tz_aliases: Dict[Tuple[str, ...], str] = {
                (
                    "+00:00",
                    "-00:00",
                    "00:00",
                    "0000",
                    "GMT +0000",
                    "GMT +00:00",
                    "GMT -00",
                    "GMT +00",
                    "GMT -0",
                    "GMT +0",
                ): "GMT0",
                ("+01:00", "+0100", "GMT +0100", "GMT +01:00", "GMT +01", "GMT +1"): "Etc/GMT-1",
                ("+02:00", "+0200", "GMT +0200", "GMT +02:00", "GMT +02", "GMT +2"): "Etc/GMT-2",
                ("+03:00", "+0300", "GMT +0300", "GMT +03:00", "GMT +03", "GMT +3"): "Etc/GMT-3",
                ("+04:00", "+0400", "GMT +0400", "GMT +04:00", "GMT +04", "GMT +4"): "Etc/GMT-4",
                ("+05:00", "+0500", "GMT +0500", "GMT +05:00", "GMT +05", "GMT +5"): "Etc/GMT-5",
                ("+06:00", "+0600", "GMT +0600", "GMT +06:00", "GMT +06", "GMT +6"): "Etc/GMT-6",
                ("+07:00", "+0700", "GMT +0700", "GMT +07:00", "GMT +07", "GMT +7"): "Etc/GMT-7",
                ("+08:00", "+0800", "GMT +0800", "GMT +08:00", "GMT +08", "GMT +8"): "Etc/GMT-8",
                ("+09:00", "+0900", "GMT +0900", "GMT +09:00", "GMT +09", "GMT +9"): "Etc/GMT-9",
                ("+10:00", "+1000", "GMT +1000", "GMT +10:00", "GMT +10"): "Etc/GMT-10",
                ("+11:00", "+1100", "GMT +1100", "GMT +11:00", "GMT +11"): "Etc/GMT-11",
                ("+12:00", "+1200", "GMT +1200", "GMT +12:00", "GMT +12"): "Etc/GMT-12",
                ("-01:00", "-0100", "GMT -0100", "GMT -01:00", "GMT -01", "GMT -1"): "Etc/GMT+1",
                ("-02:00", "-0200", "GMT -0200", "GMT -02:00", "GMT -02", "GMT -2"): "Etc/GMT+2",
                ("-03:00", "-0300", "GMT -0300", "GMT -03:00", "GMT -03", "GMT -3"): "Etc/GMT+3",
                ("-04:00", "-0400", "GMT -0400", "GMT -04:00", "GMT -04", "GMT -4"): "Etc/GMT+4",
                ("-05:00", "-0500", "GMT -0500", "GMT -05:00", "GMT -05", "GMT -5"): "Etc/GMT+5",
                ("-06:00", "-0600", "GMT -0600", "GMT -06:00", "GMT -06", "GMT -6"): "Etc/GMT+6",
                ("-07:00", "-0700", "GMT -0700", "GMT -07:00", "GMT -07", "GMT -7"): "Etc/GMT+7",
                ("-08:00", "-0800", "GMT -0800", "GMT -08:00", "GMT -08", "GMT -8"): "Etc/GMT+8",
                ("-09:00", "-0900", "GMT -0900", "GMT -09:00", "GMT -09", "GMT -9"): "Etc/GMT+9",
                ("-10:00", "-1000", "GMT -1000", "GMT -10:00", "GMT -10"): "Etc/GMT+10",
                ("-11:00", "-1100", "GMT -1100", "GMT -11:00", "GMT -11"): "Etc/GMT+11",
                ("-12:00", "-1200", "GMT -1200", "GMT -12:00", "GMT -12"): "Etc/GMT+12",
            }
            try:
                try:
                    timezone = [
                        v
                        for k, v in tz_aliases.items()
                        if timezone in k or timezone.replace(" ", "") in [x.replace(" ", "") for x in k]
                    ][0]
                except IndexError:
                    pass
                pytz.timezone(timezone or "")
            except Exception as e:
                raise Exception("Unknown timezone: {}".format(timezone)) from e

        return timezone

    @classmethod
    async def start_schedule_loop(
        cls,
        obj: Any,
        context: Dict,
        handler: Callable,
        func: Callable,
        interval: Optional[Union[str, int]] = None,
        timestamp: Optional[str] = None,
        timezone: Optional[str] = None,
        immediately: Optional[bool] = False,
    ) -> None:
        timezone = cls.get_timezone(timezone)

        if not cls.close_waiter:
            cls.close_waiter = asyncio.Future()
        stop_waiter: asyncio.Future = asyncio.Future()
        start_waiter: asyncio.Future = asyncio.Future()

        async def schedule_loop() -> None:
            sleep_task: asyncio.Future
            current_time = time.time()
            max_sleep_time = 300

            try:
                sleep_task = asyncio.ensure_future(asyncio.sleep(10))
                await asyncio.wait([sleep_task, start_waiter], return_when=asyncio.FIRST_COMPLETED)
                if not sleep_task.done():
                    sleep_task.cancel()
                else:
                    logging.getLogger("transport.schedule").warning(
                        "Scheduled loop for function '{}' cannot start yet - start waiter not done for 10 seconds".format(
                            func.__name__
                        )
                    )
                    sleep_task = asyncio.ensure_future(asyncio.sleep(110))
                    await asyncio.wait([sleep_task, start_waiter], return_when=asyncio.FIRST_COMPLETED)
                    if not sleep_task.done():
                        sleep_task.cancel()
                    else:
                        logging.getLogger("transport.schedule").warning(
                            "Scheduled loop for function '{}' cannot start yet - start waiter not done for 120 seconds".format(
                                func.__name__
                            )
                        )
                        logging.getLogger("exception").exception("Scheduled loop not started for 120 seconds")

                await asyncio.sleep(0.1)
                await start_waiter

                if not cls.close_waiter or cls.close_waiter.done():
                    logging.getLogger("transport.schedule").info(
                        "Scheduled loop for function '{}' never started before service termination".format(
                            func.__name__
                        )
                    )
                else:
                    ts0 = cls.next_call_at(current_time, interval, timestamp, timezone)
                    for _ in range(10):
                        ts1 = cls.next_call_at(ts0, interval, timestamp, timezone)
                        ts2 = cls.next_call_at(ts1, interval, timestamp, timezone)
                        if int(ts2 - ts1) // 2 < max_sleep_time:
                            max_sleep_time = int(ts2 - ts1) // 2
                    max_sleep_time = min(max(max_sleep_time, 10), 300)
            except (Exception, asyncio.CancelledError) as e:
                logging.getLogger("exception").exception("Uncaught exception: {}".format(str(e)))

                await asyncio.sleep(0.1)
                await start_waiter

                if not cls.close_waiter or cls.close_waiter.done():
                    logging.getLogger("transport.schedule").info(
                        "Scheduled loop for function '{}' never started before service termination".format(
                            func.__name__
                        )
                    )

            next_call_at = None
            prev_call_at = None
            tasks: List = []
            too_many_tasks = False
            threshold = 20
            run_immediately = immediately

            while cls.close_waiter and not cls.close_waiter.done():
                try:
                    if not run_immediately:
                        last_time = current_time
                        actual_time = time.time()
                        current_time = last_time + 1 if int(last_time + 1) < int(actual_time) else actual_time

                        if next_call_at is None:
                            next_call_at = cls.next_call_at(current_time, interval, timestamp, timezone)
                            if prev_call_at and prev_call_at == next_call_at:
                                if int(last_time + 60) < int(actual_time):
                                    logging.getLogger("transport.schedule").warning(
                                        "Scheduled tasks for function '{}' is out of time sync and may not run".format(
                                            func.__name__
                                        )
                                    )
                                    logging.getLogger("exception").exception("Scheduled task loop out of sync")

                                next_call_at = None
                                await asyncio.sleep(1)
                                continue
                        sleep_diff = int(current_time + 1) - actual_time + 0.001
                        if next_call_at > time.time() + 8:
                            sleep_diff = int((next_call_at - time.time()) / 3)
                        if sleep_diff >= max_sleep_time:
                            sleep_diff = int(max_sleep_time - random.random() * 5)
                        if sleep_diff >= 2:
                            sleep_task = asyncio.ensure_future(asyncio.sleep(sleep_diff))
                            await asyncio.wait([sleep_task, cls.close_waiter], return_when=asyncio.FIRST_COMPLETED)
                            if not sleep_task.done():
                                sleep_task.cancel()
                            current_time = time.time()
                        else:
                            await asyncio.sleep(sleep_diff)
                        if next_call_at > time.time():
                            continue
                    run_immediately = False
                    if cls.close_waiter.done():
                        continue
                    prev_call_at = next_call_at
                    next_call_at = None

                    tasks = [task for task in tasks if not task.done()]

                    if len(tasks) >= 20:
                        if not too_many_tasks and len(tasks) >= threshold:
                            too_many_tasks = True
                            logging.getLogger("transport.schedule").warning(
                                "Too many scheduled tasks ({}) for function '{}'".format(threshold, func.__name__)
                            )
                            threshold = threshold * 2
                        await asyncio.sleep(1)
                        current_time = time.time()
                        next_call_at = cls.next_call_at(current_time + 10, interval, timestamp, timezone)
                        continue
                    if too_many_tasks and len(tasks) >= 15:
                        await asyncio.sleep(1)
                        current_time = time.time()
                        next_call_at = cls.next_call_at(current_time + 10, interval, timestamp, timezone)
                        continue
                    if too_many_tasks and len(tasks) < 15:
                        logging.getLogger("transport.schedule").info(
                            "Tasks within threshold for function '{}' - resumed".format(func.__name__)
                        )
                        threshold = 20
                    too_many_tasks = False

                    current_time = time.time()
                    task = asyncio.ensure_future(handler())
                    if hasattr(task, "set_name"):
                        getattr(task, "set_name")(
                            "{} : {}".format(
                                func.__name__, datetime.datetime.utcfromtimestamp(current_time).isoformat()
                            )
                        )
                    tasks.append(task)
                except (Exception, asyncio.CancelledError) as e:
                    logging.getLogger("exception").exception("Uncaught exception: {}".format(str(e)))
                    await asyncio.sleep(1)
                except BaseException as e:
                    logging.getLogger("exception").exception("Uncaught exception: {}".format(str(e)))
                    await asyncio.sleep(1)

            if tasks:
                task_waiter = asyncio.ensure_future(asyncio.wait(tasks))
                sleep_task = asyncio.ensure_future(asyncio.sleep(2))
                await asyncio.wait([sleep_task, task_waiter], return_when=asyncio.FIRST_COMPLETED)
                if not sleep_task.done():
                    sleep_task.cancel()
                for task in tasks:
                    if task.done():
                        continue
                    task_name = getattr(task, "get_name")() if hasattr(task, "get_name") else func.__name__
                    logging.getLogger("transport.schedule").warning(
                        "Awaiting task '{}' to finish execution".format(task_name)
                    )

                while not task_waiter.done():
                    sleep_task = asyncio.ensure_future(asyncio.sleep(10))
                    await asyncio.wait([sleep_task, task_waiter], return_when=asyncio.FIRST_COMPLETED)
                    if not sleep_task.done():
                        sleep_task.cancel()
                    for task in tasks:
                        if task.done():
                            continue
                        task_name = getattr(task, "get_name")() if hasattr(task, "get_name") else func.__name__
                        logging.getLogger("transport.schedule").warning(
                            "Still awaiting task '{}' to finish execution".format(task_name)
                        )

            if not stop_waiter.done():
                stop_waiter.set_result(None)

        loop: Any = asyncio.get_event_loop()

        stop_method = getattr(obj, "_stop_service", None)

        async def stop_service(*args: Any, **kwargs: Any) -> None:
            if cls.close_waiter and not cls.close_waiter.done():
                cls.close_waiter.set_result(None)

                if not start_waiter.done():
                    start_waiter.set_result(None)
                await stop_waiter
                if stop_method:
                    await stop_method(*args, **kwargs)
            else:
                await stop_waiter
                if stop_method:
                    await stop_method(*args, **kwargs)

        setattr(obj, "_stop_service", stop_service)

        started_method = getattr(obj, "_started_service", None)

        async def started_service(*args: Any, **kwargs: Any) -> None:
            if started_method:
                await started_method(*args, **kwargs)
            if not start_waiter.done():
                start_waiter.set_result(None)

        setattr(obj, "_started_service", started_service)

        loop.create_task(schedule_loop())

    @classmethod
    async def start_scheduler(cls, obj: Any, context: Dict) -> Optional[Callable]:
        if context.get("_schedule_loop_started"):
            return None
        context["_schedule_loop_started"] = True

        set_execution_context(
            {
                "scheduled_functions_enabled": True,
                "scheduled_functions_current_tasks": 0,
                "scheduled_functions_total_tasks": 0,
            }
        )

        async def _schedule() -> None:
            cls.close_waiter = asyncio.Future()

            for interval, timestamp, timezone, immediately, func, handler in context.get(
                "_schedule_scheduled_functions", []
            ):
                cls.next_call_at(
                    time.time(), interval, timestamp, cls.get_timezone(timezone)
                )  # test provided interval/timestamp on init

            for interval, timestamp, timezone, immediately, func, handler in context.get(
                "_schedule_scheduled_functions", []
            ):
                await cls.start_schedule_loop(obj, context, handler, func, interval, timestamp, timezone, immediately)

        return _schedule


__schedule = Scheduler.decorator(Scheduler.schedule_handler)
__scheduler = Scheduler.decorator(Scheduler.schedule_handler)

__heartbeat = Scheduler.decorator(Scheduler.schedule_handler_with_interval(1))
__every_second = Scheduler.decorator(Scheduler.schedule_handler_with_interval(1))

__minutely = Scheduler.decorator(Scheduler.schedule_handler_with_interval("minutely"))
__hourly = Scheduler.decorator(Scheduler.schedule_handler_with_interval("hourly"))
__daily = Scheduler.decorator(Scheduler.schedule_handler_with_interval("daily"))
__monthly = Scheduler.decorator(Scheduler.schedule_handler_with_interval("monthly"))


def schedule(
    interval: Optional[Union[str, int]] = None,
    timestamp: Optional[str] = None,
    timezone: Optional[str] = None,
    immediately: Optional[bool] = False,
) -> Callable:
    return cast(
        Callable, __schedule(interval=interval, timestamp=timestamp, timezone=timezone, immediately=immediately)
    )


def scheduler(
    interval: Optional[Union[str, int]] = None,
    timestamp: Optional[str] = None,
    timezone: Optional[str] = None,
    immediately: Optional[bool] = False,
) -> Callable:
    return cast(
        Callable, __scheduler(interval=interval, timestamp=timestamp, timezone=timezone, immediately=immediately)
    )


def heartbeat(func: Optional[Callable] = None) -> Callable:
    return cast(Callable, __heartbeat(func))


def every_second(func: Optional[Callable] = None) -> Callable:
    return cast(Callable, __every_second(func))


def minutely(func: Optional[Callable] = None) -> Callable:
    return cast(Callable, __minutely(func))


def hourly(func: Optional[Callable] = None) -> Callable:
    return cast(Callable, __hourly(func))


def daily(func: Optional[Callable] = None) -> Callable:
    return cast(Callable, __daily(func))


def monthly(func: Optional[Callable] = None) -> Callable:
    return cast(Callable, __monthly(func))



import asyncio
import inspect
import time
from contextlib import asynccontextmanager
from typing import Any, AsyncIterator, Dict, Optional, cast

import aiobotocore
import aiobotocore.client
import aiobotocore.config
import aiobotocore.session
import aiohttp
import aiohttp.client_exceptions
import botocore
import botocore.exceptions

MAX_POOL_CONNECTIONS = 50
CONNECT_TIMEOUT = 8
READ_TIMEOUT = 35
CLIENT_CREATION_TIME_LOCK = 45


class ClientConnector(object):
    __slots__ = (
        "clients",
        "credentials",
        "client_creation_lock_time",
        "aliases",
        "locks",
        "close_waiter",
    )

    clients: Dict[str, Optional[aiobotocore.client.AioBaseClient]]
    credentials: Dict[str, Dict]
    client_creation_lock_time: Dict[str, float]
    aliases: Dict[str, str]
    locks: Dict[str, asyncio.Lock]
    close_waiter: Optional[asyncio.Future]

    def __init__(self) -> None:
        self.clients = {}
        self.credentials = {}
        self.aliases = {}
        self.client_creation_lock_time = {}
        self.locks = {}
        self.close_waiter = None

    def setup_credentials(self, alias_name: str, credentials: Dict) -> None:
        self.credentials[alias_name] = credentials

    def get_client(self, alias_name: str) -> Optional[aiobotocore.client.AioBaseClient]:
        return self.clients.get(alias_name)

    def get_lock(self, alias_name: str) -> asyncio.Lock:
        if alias_name not in self.locks:
            self.locks[alias_name] = asyncio.Lock()
        return self.locks[alias_name]

    async def create_client(
        self, alias_name: Optional[str] = None, credentials: Optional[Dict] = None, service_name: Optional[str] = None
    ) -> aiobotocore.client.AioBaseClient:
        if service_name is None and alias_name is not None:
            service_name = alias_name
        if alias_name is None and service_name is not None:
            alias_name = service_name

        alias_name = str(alias_name)
        service_name = str(service_name)

        exc_iteration_count = 0
        while self.close_waiter:
            if self.close_waiter.done():
                self.close_waiter = None
            else:
                try:
                    await self.close_waiter
                except RuntimeError:
                    exc_iteration_count += 1
                    if exc_iteration_count >= 100:
                        raise
                    await asyncio.sleep(0.1)

        async with self.get_lock(alias_name):
            client = self.get_client(alias_name)
            if self.client_creation_lock_time.get(alias_name, 0) + CLIENT_CREATION_TIME_LOCK > time.time():
                if client:
                    return client
            self.client_creation_lock_time[alias_name] = time.time() if client else 0

            if not credentials:
                credentials = self.credentials.get(alias_name, {})
            else:
                self.credentials[alias_name] = credentials
            if not credentials:
                credentials = {}

            self.aliases[alias_name] = service_name

            session = aiobotocore.session.get_session()
            config = aiobotocore.config.AioConfig(
                connect_timeout=CONNECT_TIMEOUT, read_timeout=READ_TIMEOUT, max_pool_connections=MAX_POOL_CONNECTIONS
            )

            create_client_func = session._create_client if hasattr(session, "_create_client") else session.create_client
            client_value = create_client_func(service_name, config=config, **credentials)
            if inspect.isawaitable(client_value):
                client = await client_value
            else:
                client = client_value

            old_client = self.get_client(alias_name)
            self.clients[alias_name] = cast(aiobotocore.client.AioBaseClient, client)

            if old_client:
                try:
                    await asyncio.sleep(1)
                    task = old_client.close()
                    if getattr(task, "_coro", None):
                        task = getattr(task, "_coro")
                    await asyncio.wait([asyncio.ensure_future(task)], timeout=3)
                    await asyncio.sleep(0.25)  # SSL termination sleep
                except (Exception, RuntimeError, asyncio.CancelledError, BaseException):
                    pass

            return cast(aiobotocore.client.AioBaseClient, self.clients.get(alias_name) or client)

    async def close_client(
        self,
        alias_name: Optional[str] = None,
        client: Optional[aiobotocore.client.AioBaseClient] = None,
        fast: bool = False,
    ) -> None:
        exc_iteration_count = 0
        while self.close_waiter:
            if self.close_waiter.done():
                self.close_waiter = None
            else:
                try:
                    await self.close_waiter
                except RuntimeError:
                    exc_iteration_count += 1
                    if exc_iteration_count >= 100:
                        raise
                    await asyncio.sleep(0.1)

        if not alias_name and client:
            if client in self.clients.values():
                alias_name = [k for k, v in self.clients.items() if client == v][0]

        if not alias_name:
            return

        async with self.get_lock(alias_name):
            client = self.get_client(alias_name)
            if not client:
                return

            try:
                if not fast:
                    await asyncio.sleep(0.25)
                task = client.close()
                if getattr(task, "_coro", None):
                    task = getattr(task, "_coro")
                await asyncio.wait([asyncio.ensure_future(task)], timeout=3)
                if not fast:
                    await asyncio.sleep(0.25)  # SSL termination sleep
                else:
                    await asyncio.sleep(0)
            except (Exception, RuntimeError, asyncio.CancelledError, BaseException):
                pass

            self.clients[alias_name] = None

    async def reconnect_client(
        self, alias_name: Optional[str] = None, client: Optional[aiobotocore.client.AioBaseClient] = None
    ) -> aiobotocore.client.AioBaseClient:
        if not alias_name and client:
            if client in self.clients.values():
                alias_name = [k for k, v in self.clients.items() if client == v][0]

        service_name = self.aliases.get(alias_name or "")
        if not service_name:
            raise Exception("Client has never been created in the first place")

        return await self.create_client(alias_name, service_name=service_name)

    async def close(self, fast: bool = False) -> None:
        if self.close_waiter and not self.close_waiter.done():
            return

        clients = self.clients

        self.clients = {}
        self.credentials = {}
        self.aliases = {}
        self.client_creation_lock_time = {}
        self.locks = {}

        if not clients:
            return

        self.close_waiter = asyncio.Future()
        if not fast:
            await asyncio.sleep(0.25)

        tasks = []
        for _, client in clients.items():
            if not client:
                continue
            try:
                task = client.close()
                if getattr(task, "_coro", None):
                    task = getattr(task, "_coro")
                tasks.append(asyncio.ensure_future(task))
            except (Exception, RuntimeError, asyncio.CancelledError, BaseException):
                pass

        try:
            await asyncio.wait(tasks, timeout=3)
            if not fast:
                await asyncio.sleep(0.25)  # SSL termination sleep
            else:
                await asyncio.sleep(0)
        except (Exception, RuntimeError, asyncio.CancelledError, BaseException):
            pass

        if self.close_waiter:
            self.close_waiter.set_result(None)

    @asynccontextmanager
    async def __call__(
        self, alias_name: Optional[str] = None, credentials: Optional[Dict] = None, service_name: Optional[str] = None
    ) -> AsyncIterator[Any]:
        exc_iteration_count = 0
        while self.close_waiter and not self.close_waiter.done():
            try:
                await self.close_waiter
            except RuntimeError:
                exc_iteration_count += 1
                if exc_iteration_count >= 100:
                    raise
                await asyncio.sleep(0.1)

        client_name = alias_name or service_name or ""

        if not self.get_client(client_name):
            await self.create_client(alias_name, credentials, service_name)
        client = self.get_client(client_name)

        try:
            yield client
        except (
            botocore.exceptions.NoRegionError,
            botocore.exceptions.PartialCredentialsError,
            botocore.exceptions.NoCredentialsError,
        ):
            await self.close_client(client=client, fast=True)
            raise
        except (aiohttp.client_exceptions.ServerDisconnectedError, asyncio.TimeoutError, RuntimeError):
            await self.reconnect_client(client_name, client=client)
            raise
        except botocore.exceptions.ClientError as e:
            error_message = str(e)
            if "The security token included in the request is invalid" in error_message:
                await self.close_client(client=client, fast=True)
            raise


connector: ClientConnector = ClientConnector()



#!/usr/bin/env python
import asyncio
import getopt
import logging
import os
import sys
from typing import Dict, List, Optional, Tuple, cast

import tomodachi
from tomodachi.config import parse_config_files
from tomodachi.launcher import ServiceLauncher

try:
    if ModuleNotFoundError:
        pass
except Exception:

    class ModuleNotFoundError(ImportError):
        pass


class CLI:
    def help_command_usage(self) -> str:
        return (
            "Usage: tomodachi <command> [options] [arguments]\n"
            "\n"
            "Options:\n"
            "  -h, --help                                Show this help message and exit\n"
            "  -v, --version                             Print tomodachi version\n"
            "  --dependency-versions                     Print versions of dependencies\n"
            "\n"
            "Available commands:\n"
            "  ---\n"
            "  Command: run\n"
            "  Starts service(s) defined in the .py files specified as <service> argument(s)\n"
            "\n"
            "  $ tomodachi run <service ...> [-c <config-file ...>] [--production]\n"
            "  | --loop [auto|asyncio|uvloop]            Event loop implementation [asyncio]\n"
            "  | --production                            Disable restart on file changes\n"
            "  | -c, --config <files>                    Use configuration from JSON files\n"
            "  | -l, --log <level>, --log-level <level>  Specify log level\n"
            "\n"
            ">> Version: {}\n"
            ">> Full documentation at: https://tomodachi.dev/docs"
        ).format(tomodachi.__version__)

    def help_command(self) -> None:
        print(self.help_command_usage())
        sys.exit(2)

    def version_command(self) -> None:
        print("tomodachi {}".format(tomodachi.__version__))
        sys.exit(0)

    def dependency_versions_command(self) -> None:
        CLI.test_dependencies(fail_on_errors=False, output_versions=True, output_errors=True)
        sys.exit(0)

    @classmethod
    def test_dependencies(
        cls, fail_on_errors: bool = True, output_versions: bool = False, output_errors: bool = True
    ) -> Dict[str, Optional[str]]:
        errors = False

        aioamqp_version = ""
        aiobotocore_version = ""
        aiohttp_version = ""
        botocore_version = ""
        protobuf_version = ""
        uvloop_version = ""

        try:
            import aioamqp  # noqa  # isort:skip

            aioamqp_version = aioamqp.__version__
            if output_versions:
                print("aioamqp/{}".format(aioamqp_version))
        except ModuleNotFoundError as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'aioamqp' failed to load (error: \"{}\")".format(str(e)))
        except Exception as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'aioamqp' failed to load (error: \"{}\")".format(str(e)))
                logging.exception("")
                print("")

        try:
            import aiobotocore  # noqa  # isort:skip

            aiobotocore_version = aiobotocore.__version__
            if output_versions:
                print("aiobotocore/{}".format(aiobotocore_version))
        except ModuleNotFoundError as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'aiobotocore' failed to load (error: \"{}\")".format(str(e)))
        except Exception as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'aiobotocore' failed to load (error: \"{}\")".format(str(e)))
                logging.exception("")
                print("")

        try:
            import aiohttp  # noqa  # isort:skip

            aiohttp_version = aiohttp.__version__
            if output_versions:
                print("aiohttp/{}".format(aiohttp_version))
        except ModuleNotFoundError as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'aiohttp' failed to load (error: \"{}\")".format(str(e)))
        except Exception as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'aiohttp' failed to load (error: \"{}\")".format(str(e)))
                logging.exception("")
                print("")

        try:
            import botocore  # noqa  # isort:skip

            botocore_version = botocore.__version__
            if output_versions:
                print("botocore/{}".format(botocore_version))
        except ModuleNotFoundError as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'botocore' failed to load (error: \"{}\")".format(str(e)))
        except Exception as e:  # pragma: no cover
            errors = True
            if output_errors:
                print("Dependency failure: 'botocore' failed to load (error: \"{}\")".format(str(e)))
                logging.exception("")
                print("")

        try:
            # Optional
            import google.protobuf  # noqa  # isort:skip

            protobuf_version_ = google.protobuf.__version__
            if isinstance(protobuf_version_, bytes):
                protobuf_version = cast(bytes, protobuf_version_).decode()
            else:
                protobuf_version = str(protobuf_version_)
            if output_versions:
                print("protobuf/{}".format(protobuf_version))
        except ModuleNotFoundError:  # pragma: no cover
            pass
        except Exception:  # pragma: no cover
            pass

        try:
            # Optional
            import uvloop  # noqa  # isort:skip

            uvloop_version = uvloop.__version__
            if output_versions:
                print("uvloop/{}".format(uvloop_version))
        except ModuleNotFoundError:  # pragma: no cover
            pass
        except Exception:  # pragma: no cover
            pass

        if not errors:
            try:
                import tomodachi.helpers.logging  # noqa  # isort:skip
                import tomodachi.invoker  # noqa  # isort:skip
                import tomodachi.transport.amqp  # noqa  # isort:skip
                import tomodachi.transport.aws_sns_sqs  # noqa  # isort:skip
                import tomodachi.transport.http  # noqa  # isort:skip
                import tomodachi.transport.schedule  # noqa  # isort:skip
            except Exception as e:  # pragma: no cover
                errors = True
                if output_errors:
                    print('Dependency failure: tomodachi essentials failed to load (error: "{}")'.format(str(e)))
                    logging.exception("")
                    print("")

        if errors:
            if fail_on_errors:
                logging.getLogger("exception").warning("Unable to initialize dependencies")
                logging.getLogger("exception").warning("Error: See above exceptions and traceback")
                sys.exit(1)
            elif output_errors:
                print("There were errors - see above for exceptions and traceback")

        return {
            "aioamqp": aioamqp_version or None,
            "aiobotocore": aiobotocore_version or None,
            "aiohttp": aiohttp_version or None,
            "botocore": botocore_version or None,
            "protobuf": protobuf_version or None,
            "uvloop": uvloop_version or None,
        }

    def run_command_usage(self) -> str:
        return "Usage: tomodachi run <service ...> [-c <config-file ...>] [--loop auto|asyncio|uvloop] [--production]"

    def run_command(self, args: List[str]) -> None:
        if len(args) == 0:
            print(self.run_command_usage())
        else:
            configuration = None
            log_level = logging.INFO

            env_loop = str(os.getenv("TOMODACHI_LOOP", "")).lower() or None

            if env_loop or "--loop" in args:
                if "--loop" in args:
                    index = args.index("--loop")
                    args.pop(index)
                    value = args.pop(index).lower()

                    if env_loop and env_loop != value:
                        print("Invalid argument to --loop, '{}' differs from env TOMODACHI_LOOP".format(value))
                        sys.exit(2)
                elif env_loop:
                    value = env_loop
                else:
                    value = "auto"

                if value in ("auto", "default"):
                    pass
                elif value in ("asyncio", "aio", "async"):
                    asyncio.set_event_loop_policy(asyncio.DefaultEventLoopPolicy())
                    pass
                elif value in ("uvloop", "libuv", "uv"):
                    try:
                        import uvloop  # noqa  # isort:skip
                    except Exception:  # pragma: no cover
                        print("The 'uvloop' package needs to be installed to use uvloop event loop")
                        sys.exit(2)
                    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
                else:
                    print("Invalid argument to --loop, event loop '{}' not recognized".format(value))
                    sys.exit(2)

            if "-c" in args or "--config" in args:
                index = args.index("-c") if "-c" in args else args.index("--config")
                args.pop(index)

                config_files: List[str] = []
                while len(args) > index and args[index][0] != "-":
                    value = args.pop(index)
                    if value not in config_files:
                        config_files.append(value)

                if not len(config_files):
                    print("Missing config file on command line")
                    sys.exit(2)

                try:
                    configuration = parse_config_files(config_files)
                except FileNotFoundError as e:
                    print("Invalid config file: {}".format(str(e)))
                    sys.exit(2)
                except ValueError as e:
                    print("Invalid config file, invalid JSON format: {}".format(str(e)))
                    sys.exit(2)

            env_production = str(os.getenv("TOMODACHI_PRODUCTION", "")).lower() or None
            if env_production and env_production in ("0", "no", "none", "false"):
                env_production = None

            if env_production or "--production" in args:
                if "--production" in args:
                    index = args.index("--production")
                    args.pop(index)
                watcher = None
            else:
                cwd = os.path.realpath(os.getcwd())
                root_directories = [cwd]
                for arg in set(args):
                    if not arg.startswith("/") and not arg.startswith("~"):
                        root_directories.append(os.path.realpath(os.path.dirname(os.path.join(cwd, arg))))
                    else:
                        root_directories.append(os.path.realpath(os.path.dirname(arg)))
                for p in str(os.getenv("PYTHONPATH", "")).split(os.pathsep):
                    if not p:
                        continue
                    if not p.startswith("/") and not p.startswith("~"):
                        root_directories.append(os.path.realpath(os.path.join(cwd, p)))
                    else:
                        root_directories.append(os.path.realpath(p))

                from tomodachi.watcher import Watcher  # noqa  #  isort:skip

                watcher = Watcher(root=root_directories, configuration=configuration)

            if "-l" in args or "--log" in args or "--log-level" in args:
                index = (
                    args.index("-l")
                    if "-l" in args
                    else args.index("--log")
                    if "--log" in args
                    else args.index("--log-level")
                )
                args.pop(index)
                if len(args) > index:
                    log_level = getattr(logging, args.pop(index).upper(), None) or log_level

            logging.basicConfig(format="%(asctime)s (%(name)s): %(message)s", level=log_level)
            logging.Formatter(fmt="%(asctime)s.%(msecs).03d", datefmt="%Y-%m-%d %H:%M:%S")

            ServiceLauncher.run_until_complete(set(args), configuration, watcher)
        sys.exit(tomodachi.SERVICE_EXIT_CODE)

    def main(self, argv: List[str]) -> None:
        opts: List[Tuple[str, str]] = []
        args: List[str] = []

        try:
            opts, args = getopt.getopt(
                argv, "hlvV ", ["help", "log", "version", "version", "dependency-versions", "dependencies", "deps"]
            )
        except getopt.GetoptError:
            self.help_command()
        for opt, _ in opts:
            if opt in ("-h", "--help"):
                self.help_command()
            if opt in ("-v", "-V", "--version"):
                self.version_command()
            if opt in ("--dependency-versions", "--dependencies", "--deps"):
                self.dependency_versions_command()
        if len(args):
            if args[0] in ("run", "start", "go"):
                self.run_command(args[1:])
        self.help_command()


def cli_entrypoint(argv: Optional[List[str]] = None) -> None:
    if argv is None:
        argv = sys.argv
        if argv[0].endswith("pytest"):
            argv = ["tomodachi"]

    CLI().main(argv[1:])
